import logging

import pandas as pd

from . import Predictor

_logger = logging.getLogger(__name__)


class Bias(Predictor):
    """
    A user-item bias rating prediction algorithm.  This implements the following
    predictor algorithm:

    .. math::
       s(u,i) = \\mu + b_i + b_u

    where :math:`\\mu` is the global mean rating, :math:`b_i` is item bias, and
    :math:`b_u` is the user bias.  With the provided damping values
    :math:`\\beta_{\\mathrm{u}}` and :math:`\\beta_{\\mathrm{i}}`, they are computed
    as follows:

    .. math::
       \\begin{align*}
       \\mu & = \\frac{\\sum_{r_{ui} \\in R} r_{ui}}{|R|} &
       b_i & = \\frac{\\sum_{r_{ui} \\in R_i} (r_{ui} - \\mu)}{|R_i| + \\beta_{\\mathrm{i}}} &
       b_u & = \\frac{\\sum_{r_{ui} \\in R_u} (r_{ui} - \\mu - b_i)}{|R_u| + \\beta_{\\mathrm{u}}}
       \\end{align*}

    The damping values can be interpreted as the number of default (mean) ratings to assume
    *a priori* for each user or item, damping low-information users and items towards a mean instead
    of permitting them to take on extreme values based on few ratings.

    Args:
        items: whether to compute item biases
        users: whether to compute user biases
        damping(number or tuple):
            Bayesian damping to apply to computed biases.  Either a number, to
            damp both user and item biases the same amount, or a (user,item) tuple
            providing separate damping values.

    Attributes:
        mean_(double): The global mean rating.
        item_offsets_(pandas.Series): The item offsets (:math:`b_i` values)
        user_offsets_(pandas.Series): The item offsets (:math:`b_u` values)
    """

    def __init__(self, items=True, users=True, damping=0.0):
        self.items = items
        self.users = users
        if isinstance(damping, tuple):
            self.damping = damping
            self.user_damping, self.item_damping = damping
        else:
            self.damping = damping
            self.user_damping = damping
            self.item_damping = damping

        if self.user_damping < 0:
            raise ValueError("user damping must be non-negative")
        if self.item_damping < 0:
            raise ValueError("item damping must be non-negative")

    def fit(self, ratings, **kwargs):
        """
        Train the bias model on some rating data.

        Args:
            ratings (DataFrame): a data frame of ratings. Must have at least `user`,
                                 `item`, and `rating` columns.

        Returns:
            Bias: the fit bias object.
        """

        _logger.info('building bias model for %d ratings', len(ratings))
        self.mean_ = ratings.rating.mean()
        _logger.info('global mean: %.3f', self.mean_)
        nrates = ratings.assign(rating=lambda df: df.rating - self.mean_)

        if self.items:
            group = nrates.groupby('item').rating
            self.item_offsets_ = self._mean(group, self.item_damping)
            self.item_offsets_.name = 'i_off'
            _logger.info('computed means for %d items', len(self.item_offsets_))
        else:
            self.item_offsets_ = None

        if self.users:
            if self.item_offsets_ is not None:
                nrates = nrates.join(pd.DataFrame(self.item_offsets_), on='item', how='inner')
                nrates = nrates.assign(rating=lambda df: df.rating - df.i_off)

            self.user_offsets_ = self._mean(nrates.groupby('user').rating, self.user_damping)
            self.user_offsets_.name = 'u_off'
            _logger.info('computed means for %d users', len(self.user_offsets_))
        else:
            self.user_offsets_ = None

        return self

    def transform(self, ratings, *, indexes=False):
        """
        Transform ratings by removing the bias term.  This method does *not* recompute
        user (or item) biases based on these ratings, but rather uses the biases that
        were estimated with :meth:`fit`.

        Args:
            ratings(pandas.DataFrame):
                The ratings to transform.  Must contain at least ``user``, ``item``, and
                ``rating`` columns.
            indexes(bool):
                if ``True``, the resulting frame will include ``uidx`` and ``iidx``
                columns containing the 0-based user and item indexes for each rating.

        Returns:
            pandas.DataFrame:
                A data frame with ``rating`` transformed by subtracting
                user-item bias prediction.
        """
        rvps = ratings[['user', 'item']].copy()
        rvps['rating'] = ratings['rating'] - self.mean_
        if self.item_offsets_ is not None:
            rvps = rvps.join(self.item_offsets_, on='item', how='left')
            rvps['rating'] -= rvps['i_off'].fillna(0)
        if self.user_offsets_ is not None:
            rvps = rvps.join(self.user_offsets_, on='item', how='left')
            rvps['rating'] -= rvps['u_off'].fillna(0)
        if indexes:
            rvps['uidx'] = self.user_offsets_.index.get_indexer(rvps['user'])
            rvps['iidx'] = self.item_offsets_.index.get_indexer(rvps['item'])
        return rvps.drop(columns=['u_off', 'i_off'])

    def inverse_transform(self, ratings):
        """
        Transform ratings by removing the bias term.
        """
        rvps = ratings[['user', 'item']]
        rvps['rating'] = ratings['rating'] + self.mean_
        if self.item_offsets_ is not None:
            rvps = rvps.join(self.item_offsets_, on='item', how='left')
            rvps['rating'] += rvps['i_off'].fillna(0)
        if self.user_offsets_ is not None:
            rvps = rvps.join(self.user_offsets_, on='item', how='left')
            rvps['rating'] += rvps['u_off'].fillna(0)
        return rvps.drop(columns=['u_off', 'i_off'])

    def transform_user(self, ratings):
        """
        Transform a user's ratings by subtracting the bias model.

        Args:
            ratings(pandas.Series): The user's ratings, indexed by item.
                Must have at least `item` as index and `rating` column.

        Returns:
            pandas.Series:
                The transformed ratings and the user bias.
        """
        ratings = ratings.subtract(self.mean_)

        if self.item_offsets_ is not None:
            ioff = self.item_offsets_.reindex(ratings.index, fill_value=0)
            ratings = ratings - ioff

        u_offset = self._mean(ratings, self.user_damping)

        ratings = ratings.subtract(u_offset)
        return ratings, u_offset

    def inverse_transform_user(self, user, ratings, user_bias=None):
        """
        Un-transform a user's ratings by adding in the bias model.

        Args:
            user: The user ID.
            ratings(pandas.Series): The user's ratings, indexed by item.
            user_bias(float or None): If `None`, it looks up the user bias learned by `fit`.

        Returns:
            pandas.Series: The user's de-normalized ratings.
        """
        ratings = ratings.add(self.mean_)

        if self.item_offsets_ is not None:
            ioff = self.item_offsets_.reindex(ratings.index, fill_value=0)
            ratings = ratings + ioff

        if user_bias is not None:
            ratings = ratings.add(user_bias)
        elif self.user_offsets_ is not None and user in self.user_offsets_.index:
            ratings = ratings + self.user_offsets_.loc[user]

        return ratings

    def fit_transform(self, ratings, **kwargs):
        """
        Fit with ratings and return the training data transformed.
        """
        self.fit(ratings)
        return self.transform(ratings, **kwargs)

    def predict_for_user(self, user, items, ratings=None):
        """
        Compute predictions for a user and items.  Unknown users and items
        are assumed to have zero bias.

        Args:
            user: the user ID
            items (array-like): the items to predict
            ratings (pandas.Series): the user's ratings (indexed by item id); if
                                 provided, will be used to recompute the user's
                                 bias at prediction time.

        Returns:
            pandas.Series: scores for the items, indexed by item id.
        """

        idx = pd.Index(items)
        preds = pd.Series(self.mean_, idx)

        if self.item_offsets_ is not None:
            preds = preds + self.item_offsets_.reindex(idx, fill_value=0)

        if self.users and ratings is not None:
            uoff = ratings - self.mean_
            if self.item_offsets_ is not None:
                uoff = uoff - self.item_offsets_
            umean = uoff.mean()
            preds = preds + umean
        elif self.user_offsets_ is not None:
            umean = self.user_offsets_.get(user, 0.0)
            _logger.debug('using mean(user %s) = %.3f', user, umean)
            preds = preds + umean

        return preds

    @property
    def user_index(self):
        "Get the user index from this (fit) bias."
        return self.user_offsets_.index

    @property
    def item_index(self):
        "Get the item index from this (fit) bias."
        return self.item_offsets_.index

    def _mean(self, series, damping):
        if damping is not None and damping > 0:
            return series.sum() / (series.count() + damping)
        else:
            return series.mean()

    def __str__(self):
        return 'Bias(ud={}, id={})'.format(self.user_damping, self.item_damping)



"""
The LensKit package.
"""


from lenskit.algorithms import *  # noqa: F401,F403


class DataWarning(UserWarning):
    """
    Warning raised for detectable problems with input data.
    """
    pass



import os
import logging
import pathlib
import collections
from collections.abc import Iterable
import json
import warnings
import gzip
import pickle
from copy import copy

import pandas as pd

from ..algorithms import Predictor, Recommender
from .. import util
from ._recommend import recommend
from ._predict import predict

try:
    import fastparquet
except ImportError:
    fastparquet = None

_logger = logging.getLogger(__name__)

_AlgoRec = collections.namedtuple('_AlgoRec', [
    'algorithm',
    'attributes'
])
_DSRec = collections.namedtuple('_DSRec', [
    'dataset',
    'candidates',
    'attributes'
])


class MultiEval:
    """
    A runner for carrying out multiple evaluations, such as parameter sweeps.

    Args:
        path(str or :py:class:`pathlib.Path`):
            the working directory for this evaluation.
            It will be created if it does not exist.
        predict(bool):
            whether to generate rating predictions.
        recommend(int):
            the number of recommendations to generate per user. Any false-y value (``None``,
            ``False``, ``0``) will disable top-n. The literal value ``True`` will generate
            recommendation lists of unlimited size.
        candidates(function):
            the default candidate set generator for recommendations.  It should take the
            training data and return a candidate generator, itself a function mapping user
            IDs to candidate sets.  Pass ``None`` to use the default candidate set configured
            for each algorithm (recommended).
        save_models(bool or str):
            save individual estimated models to disk.  If ``True``, models are pickled to
            ``.pkl`` files; if ``'gzip'``, they are pickled to gzip-compressed ``.pkl.gz``
            files.
        eval_n_jobs(int or None):
            Value to pass to the ``n_jobs`` parameter in :func:`lenskit.batch.predict` and
            :func:`lenskit.batch.recommend`.
        combine(bool):
            whether to combine output; if ``False``, output will be left in separate files, if
            ``True``, it will be in a single set of files (runs, recommendations, and predictions).
    """

    def __init__(self, path, *, predict=True,
                 recommend=100, candidates=None,
                 save_models=False,
                 eval_n_jobs=None, combine=True, **kwargs):
        if eval_n_jobs is None and 'nprocs' in kwargs:
            warnings.warn('nprocs is deprecated, use eval_n_jobs', DeprecationWarning)
            eval_n_jobs = kwargs['nprocs']
        self.workdir = pathlib.Path(path)
        self.predict = predict
        self.recommend = recommend
        self.candidate_generator = candidates
        self.save_models = save_models
        self.algorithms = []
        self.datasets = []
        self.n_jobs = eval_n_jobs
        self.combine_output = combine
        self._is_flat = True

    @property
    def run_csv(self):
        return self.workdir / 'runs.csv'

    @property
    def run_file(self):
        return self.workdir / 'runs.parquet'

    @property
    def preds_file(self):
        return self.workdir / 'predictions.parquet'

    @property
    def recs_file(self):
        return self.workdir / 'recommendations.parquet'

    def add_algorithms(self, algos, attrs=[], **kwargs):
        """
        Add one or more algorithms to the run.

        Args:
            algos(algorithm or list): the algorithm(s) to add.
            attrs(list of str):
                a list of attributes to extract from the algorithm objects and include in
                the run descriptions.
            kwargs:
                additional attributes to include in the run descriptions.
        """

        if not isinstance(algos, Iterable):
            algos = [algos]

        for algo in algos:
            aa = {'AlgoClass': algo.__class__.__name__, 'AlgoStr': str(algo)}
            aa.update(kwargs)
            for an in attrs:
                aa[an] = getattr(algo, an, None)

            self.algorithms.append(_AlgoRec(algo, aa))

    def add_datasets(self, data, name=None, candidates=None, **kwargs):
        """
        Add one or more datasets to the run.

        Args:
            data:
                The input data set(s) to run. Can be one of the following:

                * A tuple of (train, test) data.
                * An iterable of (train, test) pairs, in which case the iterable
                  is not consumed until it is needed.
                * A function yielding either of the above, to defer data load
                  until it is needed.

                Data can be either data frames or paths; paths are loaded after
                detection using :py:func:`util.read_df_detect`.

            kwargs:
                additional attributes pertaining to these data sets.
        """

        attrs = {}
        if name is not None:
            attrs['DataSet'] = name
        attrs.update(kwargs)

        # special-case lists to keep multis flat
        if isinstance(data, list):
            for part, e in enumerate(data):
                self.add_datasets(e, name, candidates,
                                  Partition=part+1, **kwargs)
            return

        if not isinstance(data, tuple):
            self._is_flat = False

        self.datasets.append(_DSRec(data, candidates, attrs))

    def persist_data(self):
        """
        Persist the data for an experiment, replacing in-memory data sets with file names.
        Once this has been called, the sweep can be pickled.
        """
        self.workdir.mkdir(parents=True, exist_ok=True)
        ds2 = []
        for i, (ds, cand_f, ds_attrs) in enumerate(self._flat_datasets()):
            train, test = ds
            if isinstance(train, pd.DataFrame):
                fn = self.workdir / 'ds{}-train.parquet'.format(i+1)
                _logger.info('serializing to %s', fn)
                train.to_parquet(fn)
                train = fn
            if isinstance(test, pd.DataFrame):
                fn = self.workdir / 'ds{}-test.parquet'.format(i+1)
                _logger.info('serializing to %s', fn)
                test.to_parquet(fn)
                test = fn
            ds2.append(((train, test), cand_f, ds_attrs))
        self.datasets = ds2
        self._is_flat = True

    def _normalize_ds_entry(self, entry):
        # normalize data set to be an iterable of tuples
        ds, cand_f, attrs = entry
        if callable(ds):
            ds = ds()
        if isinstance(ds, tuple):
            yield _DSRec(ds, cand_f, attrs)
        else:
            yield from (_DSRec(dse, cand_f, dict(Partition=part+1, **attrs))
                        for (part, dse) in enumerate(ds))

    def _flat_datasets(self):
        for entry in self.datasets:
            yield from self._normalize_ds_entry(entry)

    def _read_data(self, df):
        if isinstance(df, str) or isinstance(df, pathlib.Path):
            _logger.info('reading from %s', df)
            return util.read_df_detect(df)
        else:
            return df

    def _flat_runs(self):
        for dse in self._flat_datasets():
            for arec in self.algorithms:
                yield (dse, arec)

    def run_count(self):
        "Get the number of runs in this evaluation."
        if self._is_flat:
            nds = len(self.datasets)
        else:
            _logger.warning('attempting to count runs in a non-flattened evaluation')
            nds = len(list(self._flat_datasets()))
        return nds * len(self.algorithms)

    def run(self, runs=None, *, progress=None):
        """
        Run the evaluation.

        Args:
            runs(int or set-like):
                If provided, a specific set of runs to run.  Useful for splitting
                an experiment into individual runs.  This is a set of 1-based run
                IDs, not 0-based indexes.
            progress:
                A :py:func:`tqdm.tqdm`-compatible progress function.
        """

        if runs is not None and self.combine_output:
            raise ValueError('Cannot select runs with combined output')

        if runs is not None and not isinstance(runs, Iterable):
            runs = [runs]

        self.workdir.mkdir(parents=True, exist_ok=True)

        run_id = 0
        run_data = []
        train_load = util.LastMemo(self._read_data)
        test_load = util.LastMemo(self._read_data)

        iter = self._flat_runs()
        if progress is not None:
            n = self.run_count() if self._is_flat else None
            iter = progress(iter, total=n)

        for i, (dsrec, arec) in enumerate(iter):
            run_id = i + 1
            if runs is not None and run_id not in runs:
                _logger.info('skipping deselected run %d', run_id)
                continue

            ds, cand_f, ds_attrs = dsrec
            if cand_f is None:
                cand_f = self.candidate_generator
            train, test = ds
            train = train_load(train)
            test = test_load(test)

            ds_name = ds_attrs.get('DataSet', None)
            ds_part = ds_attrs.get('Partition', None)
            cand = cand_f(train) if cand_f is not None else None

            _logger.info('starting run %d: %s on %s:%s', run_id, arec.algorithm,
                         ds_name, ds_part)
            run = self._run_algo(run_id, arec, (train, test, ds_attrs, cand))
            _logger.info('finished run %d: %s on %s:%s', run_id, arec.algorithm,
                         ds_name, ds_part)
            run_data.append(run)
            self._write_run(run, run_data)

    def _run_algo(self, run_id, arec, data):
        train, test, dsp_attrs, cand = data

        run = {'RunId': run_id}
        run.update(dsp_attrs)
        run.update(arec.attributes)

        algo, train_time = self._train_algo(arec.algorithm, train)
        run['TrainTime'] = train_time
        self._save_model(run_id, algo)

        preds, pred_time = self._predict(run_id, algo, test)
        run['PredTime'] = pred_time
        self._write_results('predictions', preds, run_id)

        recs, rec_time = self._recommend(run_id, algo, test, cand)
        run['RecTime'] = rec_time
        self._write_results('recommendations', recs, run_id)

        return run

    def _save_model(self, run_id, algo):
        "Save a model to disk"
        base = 'model-{}'.format(run_id)
        base = self.workdir / base
        if not self.save_models:
            return
        elif self.save_models == 'gzip':
            with gzip.open(os.fspath(base.with_suffix('.pkl.gz')), 'wb') as f:
                pickle.dump(algo, f)
        else:
            with base.with_suffix('.pkl').open('wb') as f:
                pickle.dump(algo, f)

    def _train_algo(self, algo, train):
        watch = util.Stopwatch()
        # clone the algorithm in case some cannot be reused
        clone = util.clone(algo)
        if self.recommend and not isinstance(clone, Recommender):
            _logger.info('adapting %s into a recommender', clone)
            clone = Recommender.adapt(clone)

        _logger.info('training algorithm %s on %d ratings', algo, len(train))
        clone.fit(train)
        watch.stop()
        _logger.info('trained algorithm %s in %s', algo, watch)
        return clone, watch.elapsed()

    def _predict(self, rid, algo, test):
        if not self.predict:
            return None, None
        if not isinstance(algo, Predictor):
            return None, None
        if 'rating' not in test.columns:
            _logger.info('test data has no ratings, skipping prediction')
            return None, None

        watch = util.Stopwatch()
        _logger.info('generating %d predictions for %s', len(test), algo)
        preds = predict(algo, test, n_jobs=self.n_jobs)
        watch.stop()
        _logger.info('generated predictions in %s', watch)
        preds['RunId'] = rid
        preds = preds[['RunId', 'user', 'item', 'rating', 'prediction']]
        return preds, watch.elapsed()

    def _recommend(self, rid, algo, test, candidates):
        if not self.recommend:  # if recommend is any false-y val (0, None, False), turn off recs
            return None, None
        elif self.recommend is True:  # special value True means unlimited
            nrecs = None
        else:  # recommend has rec size
            nrecs = self.recommend

        watch = util.Stopwatch()
        users = test.user.unique()
        _logger.info('generating recommendations for %d users for %s', len(users), algo)
        recs = recommend(algo, users, nrecs, candidates,
                         n_jobs=self.n_jobs)
        watch.stop()
        _logger.info('generated recommendations in %s', watch)
        recs['RunId'] = rid
        return recs, watch.elapsed()

    def _write_run(self, run, run_data):
        if self.combine_output:
            run_df = pd.DataFrame(run_data)
            # overwrite files to show progress
            run_df.to_csv(self.run_csv, index=False)
            run_df.to_parquet(self.run_file, compression=None)
        else:
            rf = self.workdir / 'run-{}.json'.format(run['RunId'])
            with rf.open('w') as f:
                json.dump(run, f)

    def _write_results(self, name, df, run_id):
        if df is None:
            return

        if self.combine_output:
            out = self.workdir / '{}.parquet'.format(name)
            _logger.info('run %d: writing results to %s', run_id, out)
            out_fn = out / 'run-{}.part'.format(run_id)
            out.mkdir(exist_ok=True, parents=True)
            df.to_parquet(out_fn, index=False)
        else:
            out = self.workdir / '{}-{}.parquet'.format(name, run_id)
            _logger.info('run %d: writing results to %s', run_id, out)
            df.to_parquet(out)

    def collect_results(self):
        """
        Collect the results from non-combined runs into combined output files.
        """

        oc = self.combine_output
        try:
            self.combine_output = True
            n = self.run_count()
            runs = (self._read_json('run-{}.json', i+1) for i in range(n))
            runs = pd.DataFrame.from_records(runs)
            runs.to_parquet(self.run_file)
            runs.to_csv(self.run_csv, index=False)

            for i in range(n):
                preds = self._read_parquet('predictions-{}.parquet', i+1)
                self._write_results('predictions', preds, i+1)
                recs = self._read_parquet('recommendations-{}.parquet', i+1)
                self._write_results('recommendations', recs, i+1)
        finally:
            self.combine_output = oc

    def _read_parquet(self, name, *args):
        fn = self.workdir / name.format(*args)
        if not fn.exists():
            _logger.warning('file %s does not exist', fn)
            return None

        return pd.read_parquet(fn)

    def _read_json(self, name, *args):
        fn = self.workdir / name.format(*args)
        if not fn.exists():
            _logger.warning('file %s does not exist', fn)
            return {}

        with fn.open('r') as f:
            return json.load(f)

    def __getstate__(self):
        if not self._is_flat:
            _logger.warning('attempting to pickle non-flattened experiment')
        state = copy(self.__dict__)
        # clone the algorithms to only pickle their parameters
        state['algorithms'] = [a._replace(algorithm=util.clone(a.algorithm))
                               for a in self.algorithms]
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)



"""
LensKit algorithms.

The `lenskit.algorithms` package contains several example algorithms for carrying out recommender
experiments.  These algorithm implementations are designed to mimic the characteristics of the
implementations provided by the original LensKit Java package.  It also provides abstract base
classes (:py:mod:`abc`) representing different algorithm capabilities.
"""

from abc import ABCMeta, abstractmethod
import inspect

__all__ = ['Algorithm', 'Recommender', 'Predictor', 'CandidateSelector']


class Algorithm(metaclass=ABCMeta):
    """
    Base class for LensKit algorithms.  These algorithms follow the SciKit design pattern
    for estimators.

    :canonical: lenskit.Algorithm
    """

    @abstractmethod
    def fit(self, ratings, **kwargs):
        """
        Train a model using the specified ratings (or similar) data.

        Args:
            ratings(pandas.DataFrame): The ratings data.
            kwargs: Additional training data the algorithm may require.  Algorithms should
                avoid using the same keyword arguments for different purposes, so that
                they can be more easily hybridized.

        Returns:
            The algorithm object.
        """
        raise NotImplementedError()

    def get_params(self, deep=True):
        """
        Get the parameters for this algorithm (as in scikit-learn).  Algorithm parameters
        should match constructor argument names.

        The default implementation returns all attributes that match a constructor parameter
        name.  It should be compatible with :py:meth:`scikit.base.BaseEstimator.get_params`
        method so that LensKit alogrithms can be cloned with :py:func:`scikit.base.clone`
        as well as :py:func:`lenskit.util.clone`.

        Returns:
            dict: the algorithm parameters.
        """
        sig = inspect.signature(self.__class__)
        names = list(sig.parameters.keys())
        params = {}
        for name in names:
            if hasattr(self, name):
                value = getattr(self, name)
                params[name] = value
                if deep and hasattr(value, 'get_params'):
                    sps = value.get_params(deep)
                    for k, sv in sps.items():
                        params[name + '__' + k] = sv

        return params


class Predictor(Algorithm, metaclass=ABCMeta):
    """
    Predicts user ratings of items.  Predictions are really estimates of the user's like or
    dislike, and the ``Predictor`` interface makes no guarantees about their scale or
    granularity.
    """

    def predict(self, pairs, ratings=None):
        """
        Compute predictions for user-item pairs.  This method is designed to be compatible with the
        general SciKit paradigm; applications typically want to use :py:meth:`predict_for_user`.

        Args:
            pairs(pandas.DataFrame): The user-item pairs, as ``user`` and ``item`` columns.
            ratings(pandas.DataFrame): user-item rating data to replace memorized data.

        Returns:
            pandas.Series: The predicted scores for each user-item pair.
        """
        if ratings is not None:
            raise NotImplementedError()

        def upred(df):
            user, = df['user'].unique()
            items = df['item']
            preds = self.predict_for_user(user, items)
            preds.name = 'prediction'
            res = df.join(preds, on='item', how='left')
            return res.prediction

        res = pairs.loc[:, ['user', 'item']].groupby('user', sort=False).apply(upred)
        res.reset_index(level='user', inplace=True, drop=True)
        res.name = 'prediction'
        return res.loc[pairs.index.values]

    @abstractmethod
    def predict_for_user(self, user, items, ratings=None):
        """
        Compute predictions for a user and items.

        Args:
            user: the user ID
            items (array-like): the items to predict
            ratings (pandas.Series):
                the user's ratings (indexed by item id); if provided, they may be used to
                override or augment the model's notion of a user's preferences.

        Returns:
            pandas.Series: scores for the items, indexed by item id.
        """
        raise NotImplementedError()


class Recommender(Algorithm, metaclass=ABCMeta):
    """
    Recommends lists of items for users.
    """

    @abstractmethod
    def recommend(self, user, n=None, candidates=None, ratings=None):
        """
        Compute recommendations for a user.

        Args:
            user: the user ID
            n(int): the number of recommendations to produce (``None`` for unlimited)
            candidates (array-like):
                The set of valid candidate items; if ``None``, a default set will be used.
                For many algorithms, this is their :py:class:`CandidateSelector`.
            ratings (pandas.Series):
                the user's ratings (indexed by item id); if provided, they may be used to
                override or augment the model's notion of a user's preferences.

        Returns:
            pandas.DataFrame:
                a frame with an ``item`` column; if the recommender also produces scores,
                they will be in a ``score`` column.
        """
        raise NotImplementedError()

    @classmethod
    def adapt(cls, algo):
        """
        Ensure that an algorithm is a :class:`Recommender`.  If it is not a recommender,
        it is wrapped in a :class:`lenskit.basic.TopN` with a default candidate selector.

        .. note::
            Since 0.6.0, since algorithms are fit directly, you should call this method
            **before** calling :meth:`Algorithm.fit`, unless you will always be passing
            explicit candidate sets to :meth:`recommend`.

        Args:
            algo(Predictor): the underlying rating predictor.
        """
        from .basic import TopN
        if isinstance(algo, Recommender):
            return algo
        else:
            return TopN(algo)


class CandidateSelector(Algorithm, metaclass=ABCMeta):
    """
    Select candidates for recommendation for a user, possibly with some
    additional ratings.

    :class:`.UnratedItemCandidateSelector` is the default and most common implementation
    of this interface.
    """

    @abstractmethod
    def candidates(self, user, ratings=None):
        """
        Select candidates for the user.

        Args:
            user:
                The user key or ID.
            ratings(pandas.Series or array-like):
                Ratings or items to use instead of whatever ratings were memorized
                for this user.  If a :py:class:`pandas.Series`, the series index
                is used; if it is another array-like it is assumed to be an array
                of items.
        """
        raise NotImplementedError()

    @staticmethod
    def rated_items(ratings):
        """
        Utility function for converting a series or array into an array of item
        IDs.  Useful in implementations of :py:meth:`candidates`.
        """
        import pandas as pd
        import numpy as np
        if isinstance(ratings, pd.Series):
            return ratings.index.values
        elif isinstance(ratings, np.ndarray):
            return ratings
        else:
            return np.array(ratings)



"""
Numba-accleratable matrix operations.
"""

import logging
import numpy as np
import numba as n
from numba import njit
try:
    from numba.experimental import jitclass
except ImportError:
    from numba import jitclass

from ..util.array import swap
from . import _CSR

_log = logging.getLogger(__name__)

_CSR64 = type('_CSR64', _CSR.__bases__, dict(_CSR.__dict__))
_CSR = jitclass({
    'nrows': n.intc,
    'ncols': n.intc,
    'nnz': n.intc,
    'rowptrs': n.intc[::1],
    'colinds': n.intc[::1],
    'values': n.float64[::1]
})(_CSR)
_CSR64 = jitclass({
    'nrows': n.intc,
    'ncols': n.intc,
    'nnz': n.int64,
    'rowptrs': n.int64[::1],
    'colinds': n.intc[::1],
    'values': n.float64[::1]
})(_CSR64)


@njit(nogil=True)
def _center_rows(csr: _CSR):
    means = np.zeros(csr.nrows)
    for i in range(csr.nrows):
        sp, ep = csr.row_extent(i)
        if sp == ep:
            continue  # empty row
        vs = csr.row_vs(i)
        m = np.mean(vs)
        means[i] = m
        csr.values[sp:ep] -= m

    return means


@njit(nogil=True)
def _unit_rows(csr: _CSR):
    norms = np.zeros(csr.nrows)
    for i in range(csr.nrows):
        sp, ep = csr.row_extent(i)
        if sp == ep:
            continue  # empty row
        vs = csr.row_vs(i)
        m = np.linalg.norm(vs)
        norms[i] = m
        csr.values[sp:ep] /= m

    return norms


@njit(nogil=True)
def _csr_align(rowinds, nrows, rowptrs, align):
    rcts = np.zeros(nrows, dtype=rowptrs.dtype)
    for r in rowinds:
        rcts[r] += 1

    rowptrs[1:] = np.cumsum(rcts)
    rpos = rowptrs[:-1].copy()

    for i in range(len(rowinds)):
        row = rowinds[i]
        pos = rpos[row]
        align[pos] = i
        rpos[row] += 1


@njit(nogil=True)
def _csr_align_inplace(shape, rows, cols, vals):
    """
    Align COO data in-place for a CSR matrix.

    Args:
        shape: the matrix shape
        rows: the matrix row indices (not modified)
        cols: the matrix column indices (**modified**)
        vals: the matrix values (**modified**)

    Returns:
        the CSR row pointers
    """
    nrows, ncols = shape
    nnz = len(rows)

    rps = np.zeros(nrows + 1, np.int64)

    for i in range(nnz):
        rps[rows[i] + 1] += 1
    for i in range(nrows):
        rps[i+1] += rps[i]

    rci = rps[:nrows].copy()

    pos = 0
    row = 0
    rend = rps[1]

    # skip to first nonempty row
    while row < nrows and rend == 0:
        row += 1
        rend = rps[row + 1]

    while pos < nnz:
        r = rows[pos]
        # swap until we have something in place
        while r != row:
            tgt = rci[r]
            # swap with the target position
            swap(cols, pos, tgt)
            if vals is not None:
                swap(vals, pos, tgt)

            # update the target start pointer
            rci[r] += 1

            # update the loop check
            r = rows[tgt]

        # now the current entry in the arrays is good
        # we need to advance to the next entry
        pos += 1
        rci[row] += 1

        # skip finished rows
        while pos == rend and pos < nnz:
            row += 1
            pos = rci[row]
            rend = rps[row+1]

    return rps


@njit
def _empty_csr(nrows, ncols, sizes):
    nnz = np.sum(sizes)
    rowptrs = np.zeros(nrows + 1, dtype=np.intc)
    for i in range(nrows):
        rowptrs[i+1] = rowptrs[i] + sizes[i]
    colinds = np.full(nnz, -1, dtype=np.intc)
    values = np.full(nnz, np.nan)
    return _CSR(nrows, ncols, nnz, rowptrs, colinds, values)


@njit
def _subset_rows(csr, begin, end):
    st = csr.rowptrs[begin]
    ed = csr.rowptrs[end]
    rps = csr.rowptrs[begin:(end+1)] - st

    cis = csr.colinds[st:ed]
    if csr.values.size == 0:
        vs = csr.values
    else:
        vs = csr.values[st:ed]
    return _CSR(end - begin, csr.ncols, ed - st, rps, cis, vs)



import logging

import pandas as pd
import numpy as np

try:
    from sklearn.decomposition import TruncatedSVD
    SKL_AVAILABLE = True
except ImportError:
    TruncatedSVD = None
    SKL_AVAILABLE = False

from . import Predictor
from .bias import Bias
from ..matrix import sparse_ratings
from ..util import Stopwatch

_log = logging.getLogger(__name__)


class BiasedSVD(Predictor):
    """
    Biased matrix factorization for implicit feedback using SciKit-Learn's SVD
    solver (:class:`sklearn.decomposition.TruncatedSVD`).  It operates by first
    computing the bias, then computing the SVD of the bias residuals.

    You'll generally want one of the iterative SVD implementations such as
    :class:`lennskit.algorithms.als.BiasedMF`; this is here primarily as an
    example and for cases where you want to evaluate a pure SVD implementation.
    """

    def __init__(self, features, *, damping=5, bias=True, algorithm='randomized'):
        if TruncatedSVD is None:
            raise ImportError('sklearn.decomposition')
        if bias is True:
            self.bias = Bias(damping=damping)
        else:
            self.bias = bias
        self.factorization = TruncatedSVD(features, algorithm=algorithm)

    def fit(self, ratings, **kwargs):
        timer = Stopwatch()
        _log.info('[%s] computing bias', timer)
        self.bias.fit(ratings)

        g_bias = self.bias.mean_
        u_bias = self.bias.user_offsets_
        i_bias = self.bias.item_offsets_

        _log.info('[%s] sparsifying and normalizing matrix', timer)
        r_mat, users, items = sparse_ratings(ratings, users=u_bias.index, items=i_bias.index)
        # global
        r_mat.values -= g_bias
        r_mat.values -= i_bias.values[r_mat.colinds]
        r_mat.values -= u_bias.values[r_mat.rowinds()]
        r_mat = r_mat.to_scipy()
        assert r_mat.shape == (len(u_bias), len(i_bias))

        _log.info('[%s] training SVD (k=%d)', timer, self.factorization.n_components)
        Xt = self.factorization.fit_transform(r_mat)
        self.user_components_ = Xt
        _log.info("finished model training in %s", timer)

    def predict_for_user(self, user, items, ratings=None):
        items = np.array(items)
        if user not in self.bias.user_offsets_.index:
            return pd.Series(np.nan, index=items)

        # Get index for user & usable items
        uidx = self.bias.user_offsets_.index.get_loc(user)
        iidx = self.bias.item_offsets_.index.get_indexer(items)
        good_iidx = iidx[iidx >= 0]

        _log.debug("reverse-transforming user %s (idx=%d)", user, uidx)
        Xt = self.user_components_[[uidx], :]
        X = self.factorization.inverse_transform(Xt)
        # restrict to usable desired items
        Xsel = X[0, good_iidx]
        # convert to output format and de-normalize
        scores = pd.Series(Xsel, index=items[iidx >= 0]).reindex(items)
        scores += self.bias.predict_for_user(user, items, ratings)
        return scores

    def __str__(self):
        return f'BiasedSVD({self.factorization})'



"""
Code to import commonly-used RecSys data sets into LensKit-compatible data frames.
"""

import os.path
from pathlib import Path
import logging

import pandas as pd
import numpy as np

_log = logging.getLogger(__name__)

__doctest_skip__ = []
if not os.path.exists('data/ml-100k'):
    __doctest_skip__.append('ML100K.*')
if not os.path.exists('data/ml-20m'):
    __doctest_skip__.append('MovieLens.tag_genome')
if not os.path.exists('data/ml-1m.*'):
    __doctest_skip__.append('ML1M.*')
if not os.path.exists('data/ml-10M100K'):
    __doctest_skip__.append('ML10M.*')
    __doctest_skip__.append('MLM.*')


def cached(prop):
    cache = '_cached_' + prop.__name__

    def getter(self):
        val = getattr(self, cache, None)
        if val is None:
            val = prop(self)
            setattr(self, cache, val)
        return val

    getter.__doc__ = prop.__doc__

    return property(getter)


class MovieLens:
    """
    Code for reading current MovieLens data sets, including ML-20M, ML-Latest, and
    ML-Latest-Small.

    Parameters:
        path(str or pathlib.Path): Path to the directory containing the data set.
    """

    def __init__(self, path='data/ml-20m'):
        self.path = Path(path)

    @cached
    def ratings(self):
        """
        The rating table.

        >>> mlsmall = MovieLens('data/ml-latest-small')
        >>> mlsmall.ratings
                user  item  rating   timestamp
        0          1    31     2.5  1260759144
        1          1  1029     3.0  1260759179
        2          1  1061     3.0  1260759182
        3          1  1129     2.0  1260759185
        4          1  1172     4.0  1260759205
        ...
        [100004 rows x 4 columns]
        """

        fn = self.path / 'ratings.csv'
        ratings = pd.read_csv(fn, dtype={
            'movieId': np.int32,
            'userId': np.int32,
            'rating': np.float64,
            'timestamp': np.int32
        })
        ratings.rename(columns={'userId': 'user', 'movieId': 'item'}, inplace=True)
        _log.debug('loaded %s, takes %d bytes', fn, ratings.memory_usage().sum())
        return ratings

    @cached
    def movies(self):
        """
        The movie table, with titles and genres.  It is indexed by movie ID.

        >>> mlsmall = MovieLens('data/ml-latest-small')
        >>> mlsmall.movies
                                                            title                                           genres
        item
        1                                        Toy Story (1995)      Adventure|Animation|Children|Comedy|Fantasy
        2                                          Jumanji (1995)                       Adventure|Children|Fantasy
        3                                 Grumpier Old Men (1995)                                   Comedy|Romance
        4                                Waiting to Exhale (1995)                             Comedy|Drama|Romance
        5                      Father of the Bride Part II (1995)                                           Comedy
        ...
        [9125 rows x 2 columns]
        """

        fn = self.path / 'movies.csv'
        movies = pd.read_csv(fn, dtype={
            'movieId': np.int32,
            'title': np.object,
            'genres': np.object
        })
        movies.rename(columns={'movieId': 'item'}, inplace=True)
        movies.set_index('item', inplace=True)
        _log.debug('loaded %s, takes %d bytes', fn, movies.memory_usage().sum())
        return movies

    @cached
    def links(self):
        """
        The movie link table, connecting movie IDs to external identifiers.  It is indexed
        by movie ID.

        >>> mlsmall = MovieLens('data/ml-latest-small')
        >>> mlsmall.links
                 imdbId  tmdbId
        item
        1        114709     862
        2        113497    8844
        3        113228   15602
        4        114885   31357
        5        113041   11862
        ...
        [9125 rows x 2 columns]
        """

        fn = self.path / 'links.csv'
        links = pd.read_csv(fn, dtype={
            'movieId': np.int32,
            'imdbId': np.int64,
            'tmdbId': pd.Int64Dtype()
        })
        links.rename(columns={'movieId': 'item'}, inplace=True)
        links.set_index('item', inplace=True)
        _log.debug('loaded %s, takes %d bytes', fn, links.memory_usage().sum())
        return links

    @cached
    def tags(self):
        """
        The tag application table, recording user-supplied tags for movies.


        >>> mlsmall = MovieLens('data/ml-latest-small')
        >>> mlsmall.tags
              user  ...   timestamp
        0       15  ...  1138537770
        1       15  ...  1193435061
        2       15  ...  1170560997
        3       15  ...  1170626366
        4       15  ...  1141391765
        ...
        [1296 rows x 4 columns]
        """

        fn = self.path / 'tags.csv'
        tags = pd.read_csv(fn, dtype={
            'movieId': np.int32,
            'userId': np.int32,
            'tag': np.object,
            'timestamp': np.int32
        })
        tags.rename(columns={'userId': 'user', 'movieId': 'item'}, inplace=True)
        _log.debug('loaded %s, takes %d bytes', fn, tags.memory_usage().sum())
        return tags

    @cached
    def tag_genome(self):
        """
        The tag genome table, recording inferred item-tag relevance scores.  This gets returned
        as a wide Pandas data frame, with rows indexed by item ID.

        >>> ml20m = MovieLens('data/ml-20m')
        >>> ml20m.tag_genome
        tag         007  007 (series)  18th century  ...     wwii   zombie  zombies
        item                                         ...
        1       0.02500       0.02500       0.05775  ...  0.03625  0.07775  0.02300
        2       0.03975       0.04375       0.03775  ...  0.01475  0.09025  0.01875
        3       0.04350       0.05475       0.02800  ...  0.01950  0.09700  0.01850
        4       0.03725       0.03950       0.03675  ...  0.01525  0.06450  0.01300
        5       0.04200       0.05275       0.05925  ...  0.01675  0.10750  0.01825
        ...
        [10381 rows x 1128 columns]
        """

        fn = self.path / 'genome-scores.csv'
        tags = pd.read_csv(self.path / 'genome-tags.csv')
        tags = tags.set_index('tagId')
        tags = tags['tag'].astype('category')
        genome = pd.read_csv(fn, dtype={
            'movieId': np.int32,
            'tagId': np.int32,
            'relevance': np.float64
        })
        genome.rename(columns={'userId': 'user', 'movieId': 'item'}, inplace=True)
        genome = genome.join(tags, on='tagId')
        genome = genome.pivot(index='item', columns='tag', values='relevance')
        _log.debug('loaded %s, takes %d bytes', fn, genome.memory_usage().sum())
        return genome


class ML100K:
    """
    The MovieLens 100K data set.  This older data set is in a different format from
    the more current data sets loaded by :class:`MovieLens`.
    """

    def __init__(self, path='data/ml-100k'):
        self.path = Path(path)

    @property
    def available(self):
        "Query whether the data set exists."
        return (self.path / 'u.data').exists()

    @cached
    def ratings(self):
        """
        Return the rating data (from ``u.data``).

        >>> ml = ML100K()
        >>> ml.ratings
               user  item  rating  timestamp
        0       196   242     3.0  881250949
        1       186   302     3.0  891717742
        2        22   377     1.0  878887116
        3       244    51     2.0  880606923
        4       166   346     1.0  886397596
        ...
        [100000 rows x 4 columns]
        """
        fn = self.path / 'u.data'
        ratings = pd.read_csv(fn, sep='\t', header=None,
                              names=['user', 'item', 'rating', 'timestamp'],
                              dtype={'user': np.int32, 'item': np.int32,
                                     'rating': np.float32, 'timestamp': np.int32})
        _log.debug('loaded %s', fn)
        return ratings

    @cached
    def users(self):
        """
        Return the user data (from ``u.user``).

        >>> ml = ML100K()
        >>> ml.users
              age gender     occupation     zip
        user
        1      24      M     technician   85711
        2      53      F          other   94043
        3      23      M         writer   32067
        4      24      M     technician   43537
        5      33      F          other   15213
        ...
        [943 rows x 4 columns]
        """
        fn = self.path / 'u.user'
        users = pd.read_csv(fn, sep='|', header=None,
                            names=['user', 'age', 'gender', 'occupation', 'zip'],
                            dtype={'user': np.int32, 'age': np.int8,
                                   'occupation': 'category'})
        _log.debug('loaded %s', fn)
        return users.set_index('user')

    @cached
    def movies(self):
        """
        Return the user data (from ``u.user``).

        >>> ml = ML100K()
        >>> ml.movies
                                                  title      release  ...  War Western
        item                                                          ...
        1                              Toy Story (1995)  01-Jan-1995  ...    0       0
        2                              GoldenEye (1995)  01-Jan-1995  ...    0       0
        3                             Four Rooms (1995)  01-Jan-1995  ...    0       0
        4                             Get Shorty (1995)  01-Jan-1995  ...    0       0
        5                                Copycat (1995)  01-Jan-1995  ...    0       0
        ...
        [1682 rows x 23 columns]
        """
        fn = self.path / 'u.item'
        genres = [
            'unknown', 'Action', 'Adventure', 'Animation',
            "Children's", 'Comedy', 'Crime', 'Documentary', 'Drama', 'Fantasy',
            'Film-Noir', 'Horror', 'Musical', 'Mystery', 'Romance', 'Sci-Fi',
            'Thriller', 'War', 'Western'
        ]
        items = pd.read_csv(fn, sep='|', header=None, encoding='latin1',
                            names=['item', 'title', 'release', 'vidrelease', 'imdb'] + genres)
        _log.debug('loaded %s', fn)
        return items.set_index('item')


class MLM:
    """
    Base classes for ML1M and ML10M.
    """

    def __init__(self, path):
        self.path = Path(path)

    @cached
    def ratings(self):
        """
        Return the rating data (from ``ratings.dat``).

        >>> ml = ML10M()
        >>> ml.ratings
                   user  item  rating  timestamp
        0             1   122     5.0  838985046
        1             1   185     5.0  838983525
        2             1   231     5.0  838983392
        3             1   292     5.0  838983421
        4             1   316     5.0  838983392
        ...
        [10000054 rows x 4 columns]
        """
        fn = self.path / 'ratings.dat'
        ratings = pd.read_csv(fn, sep=':', header=None,
                              names=['user', '_ui', 'item', '_ir', 'rating', '_rt', 'timestamp'],
                              usecols=[0, 2, 4, 6],
                              dtype={'user': np.int32, 'item': np.int32,
                                     'rating': np.float32, 'timestamp': np.int32})
        _log.debug('loaded %s', fn)
        return ratings

    @cached
    def movies(self):
        """
        Return the movie data (from ``movies.dat``).  Indexed by movie ID.

        >>> ml = ML10M()
        >>> ml.movies
                                                            title                                           genres
        item
        1                                        Toy Story (1995)      Adventure|Animation|Children|Comedy|Fantasy
        2                                          Jumanji (1995)                       Adventure|Children|Fantasy
        3                                 Grumpier Old Men (1995)                                   Comedy|Romance
        4                                Waiting to Exhale (1995)                             Comedy|Drama|Romance
        5                      Father of the Bride Part II (1995)                                           Comedy
        ...
        [10681 rows x 2 columns]
        """
        fn = self.path / 'movies.dat'
        movies = pd.read_csv(fn, sep=':', header=None,
                             names=['item', '_ir', 'title', '_tg', 'genres'],
                             usecols=[0, 2, 4],
                             dtype={'item': np.int32})
        movies.set_index('item', inplace=True)
        _log.debug('loaded %s', fn)
        return movies


class ML10M(MLM):
    """
    MovieLens 10M100K data set.
    """
    def __init__(self, path='data/ml-10M100K'):
        super().__init__(path)


class ML1M(MLM):
    """
    MovieLens 1M data set.

    .. note::
       Some documentation examples use ML-10M100K; that is because this class shares implementation
       with the 10M data set.
    """

    def __init__(self, path='data/ml-1m'):
        super().__init__(path)

    @cached
    def users(self):
        """
        Return the movie data (from ``users.dat``).  Indexed by user ID.

        >>> ml = ML1M()
        >>> ml.users
             gender  age    zip
        user
        1         F    1  48067
        2         M   56  70072
        3         M   25  55117
        4         M   45  02460
        5         M   25  55455
        ...
        [6040 rows x 3 columns]
        """
        fn = self.path / 'users.dat'
        users = pd.read_csv(fn, sep=':', header=None,
                            names=['user', '_ug', 'gender', '_ga', 'age',
                                   '_ao', 'occupation', '_oz', 'zip'],
                            usecols=[0, 2, 4, 8],
                            dtype={'user': np.int32, 'gender': 'category', 'age': np.int8,
                                   'timestamp': np.int32})
        users.set_index('user', inplace=True)
        _log.debug('loaded %s', fn)
        return users



from logging import getLogger
from lenskit import util

try:
    import tensorflow as tf
except ImportError:
    tf = None

_log = getLogger(__name__)


def have_usable_tensorflow():
    if tf is None:
        return False
    elif tf.__version__ < '2':
        return False
    else:
        return True


def check_tensorflow():
    if tf is None:
        raise ImportError('tensorflow')
    elif not have_usable_tensorflow():
        raise RuntimeError('TensorFlow not usable, too old?')


def init_tf_rng(spec):
    if spec is None:
        return

    seed = util.random.rng_seed(spec)
    seed, = seed.generate_state(1)
    tf.random.set_seed(seed)


def make_graph(rng_spec=None):
    "Construct a TensorFlow graph (with an optional random seed)"
    rng = util.rng(rng_spec)
    graph = tf.Graph()
    graph.seed = rng.integers(2**31 - 1)
    _log.info('using effective random seed %s (from %s)', graph.seed, rng_spec)
    return graph



import logging
from collections import namedtuple

import numpy as np
from numba import njit, prange

from .bias import Bias
from .mf_common import MFPredictor
from ..matrix import sparse_ratings
from .. import util
from ..math.solve import _dposv

_logger = logging.getLogger(__name__)

PartialModel = namedtuple('PartialModel', [
    'users', 'items',
    'user_matrix', 'item_matrix'
])


@njit
def _inplace_axpy(a, x, y):
    for i in range(len(x)):
        y[i] += a * x[i]


@njit
def _rr_solve(X, xis, y, w, reg, epochs):
    """
    RR1 coordinate descent solver.

    Args:
        X(ndarray): The feature matrix.
        xis(ndarray): Row numbers in ``X`` that are rated.
        y(ndarray): Rating values corresponding to ``xis``.
        w(ndarray): Input/output vector to solve.
    """

    nd = len(w)
    Xt = X.T[:, xis]
    resid = w @ Xt
    resid *= -1.0
    resid += y

    for e in range(epochs):
        for k in range(nd):
            xk = Xt[k, :]
            num = np.dot(xk, resid) - reg * w[k]
            denom = np.dot(xk, xk) + reg
            dw = num / denom
            w[k] += dw
            _inplace_axpy(-dw, xk, resid)


@njit(parallel=True, nogil=True)
def _train_matrix_cd(mat, this: np.ndarray, other: np.ndarray, reg: float):
    """
    One half of an explicit ALS training round using coordinate descent.

    Args:
        mat: the :math:`m \\times n` matrix of ratings
        this: the :math:`m \\times k` matrix to train
        other: the :math:`n \\times k` matrix of sample features
        reg: the regularization term
    """
    nr = mat.nrows
    nf = other.shape[1]
    assert mat.ncols == other.shape[0]
    assert mat.nrows == this.shape[0]
    assert this.shape[1] == nf

    frob = 0.0

    for i in prange(nr):
        cols = mat.row_cs(i)
        if len(cols) == 0:
            continue

        vals = mat.row_vs(i)

        w = this[i, :].copy()
        _rr_solve(other, cols, vals, w, reg * len(cols), 2)
        delta = this[i, :] - w
        frob += np.dot(delta, delta)
        this[i, :] = w

    return np.sqrt(frob)


@njit(parallel=True, nogil=True)
def _train_matrix_lu(mat, this: np.ndarray, other: np.ndarray, reg: float):
    """
    One half of an explicit ALS training round using LU-decomposition on the normal
    matrices to solve the least squares problem.

    Args:
        mat: the :math:`m \\times n` matrix of ratings
        this: the :math:`m \\times k` matrix to train
        other: the :math:`n \\times k` matrix of sample features
        reg: the regularization term
    """
    nr = mat.nrows
    nf = other.shape[1]
    regI = np.identity(nf) * reg
    assert mat.ncols == other.shape[0]
    frob = 0.0

    for i in prange(nr):
        cols = mat.row_cs(i)
        if len(cols) == 0:
            continue

        vals = mat.row_vs(i)
        M = other[cols, :]
        MMT = M.T @ M
        # assert MMT.shape[0] == ctx.n_features
        # assert MMT.shape[1] == ctx.n_features
        A = MMT + regI * len(cols)
        V = M.T @ vals
        # and solve
        _dposv(A, V, True)
        delta = this[i, :] - V
        frob += np.dot(delta, delta)
        this[i, :] = V

    return np.sqrt(frob)


@njit(nogil=True)
def _train_bias_row_lu(items, ratings, other, reg):
    """
    Args:
        items(np.ndarray[i64]): the item IDs the user has rated
        ratings(np.ndarray): the user's (normalized) ratings for those items
        other(np.ndarray): the item-feature matrix
        reg(float): the regularization term
    Returns:
        np.ndarray: the user-feature vector (equivalent to V in the current LU code)
    """
    M = other[items, :]
    nf = other.shape[1]
    regI = np.identity(nf) * reg
    MMT = M.T @ M
    A = MMT + regI * len(items)

    V = M.T @ ratings
    _dposv(A, V, True)

    return V


@njit
def _cg_a_mult(OtOr, X, y, v):
    """
    Compute the multiplication Av, where A = X'X + X'yX + λ.
    """
    XtXv = OtOr @ v
    XtyXv = X.T @ (y * (X @ v))
    return XtXv + XtyXv


@njit
def _cg_solve(OtOr, X, y, w, epochs):
    """
    Use conjugate gradient method to solve the system M†(X'X + X'yX + λ)w = M†X'(y+1).
    The parameter OtOr = X'X + λ.
    """
    nf = X.shape[1]
    # compute inverse of the Jacobi preconditioner
    Ad = np.diag(OtOr).copy()
    for i in range(X.shape[0]):
        for k in range(nf):
            Ad[k] += X[i, k] * y[i] * X[i, k]

    iM = np.reciprocal(Ad)

    # compute residuals
    b = X.T @ (y + 1.0)
    r = _cg_a_mult(OtOr, X, y, w)
    r *= -1
    r += b

    # compute initial values
    z = iM * r
    p = z

    # and solve
    for i in range(epochs):
        gam = np.dot(r, z)
        Ap = _cg_a_mult(OtOr, X, y, p)
        al = gam / np.dot(p, Ap)
        _inplace_axpy(al, p, w)
        _inplace_axpy(-al, Ap, r)
        z = iM * r
        bet = np.dot(r, z) / gam
        p = z + bet * p


@njit(nogil=True)
def _implicit_otor(other, reg):
    nf = other.shape[1]
    regmat = np.identity(nf)
    regmat *= reg
    Ot = other.T
    OtO = Ot @ other
    OtO += regmat
    return OtO


@njit(parallel=True, nogil=True)
def _train_implicit_cg(mat, this: np.ndarray, other: np.ndarray, reg: float):
    "One half of an implicit ALS training round with conjugate gradient."
    nr = mat.nrows
    nc = other.shape[0]

    assert mat.ncols == nc

    OtOr = _implicit_otor(other, reg)

    frob = 0.0

    for i in prange(nr):
        cols = mat.row_cs(i)
        if len(cols) == 0:
            continue

        rates = mat.row_vs(i)

        # we can optimize by only considering the nonzero entries of Cu-I
        # this means we only need the corresponding matrix columns
        M = other[cols, :]
        # and solve
        w = this[i, :].copy()
        _cg_solve(OtOr, M, rates, w, 3)

        # update stats
        delta = this[i, :] - w
        frob += np.dot(delta, delta)

        # put back the result
        this[i, :] = w

    return np.sqrt(frob)


@njit(parallel=True, nogil=True)
def _train_implicit_lu(mat, this: np.ndarray, other: np.ndarray, reg: float):
    "One half of an implicit ALS training round."
    nr = mat.nrows
    nc = other.shape[0]
    assert mat.ncols == nc
    OtOr = _implicit_otor(other, reg)
    frob = 0.0

    for i in prange(nr):
        cols = mat.row_cs(i)
        if len(cols) == 0:
            continue

        rates = mat.row_vs(i)

        # we can optimize by only considering the nonzero entries of Cu-I
        # this means we only need the corresponding matrix columns
        M = other[cols, :]
        # Compute M^T (C_u-I) M, restricted to these nonzero entries
        MMT = (M.T.copy() * rates) @ M
        # assert MMT.shape[0] == ctx.n_features
        # assert MMT.shape[1] == ctx.n_features
        # Build the matrix for solving
        A = OtOr + MMT
        # Compute RHS - only used columns (p_ui != 0) values needed
        # Cu is rates + 1 for the cols, so just trim Ot
        y = other.T[:, cols] @ (rates + 1.0)
        # and solve
        _dposv(A, y, True)
        # assert len(uv) == ctx.n_features
        delta = this[i, :] - y
        frob += np.dot(delta, delta)
        this[i, :] = y

    return np.sqrt(frob)


@njit(nogil=True)
def _train_implicit_row_lu(items, ratings, other, otOr):
    """
    Args:
        items(np.ndarray[i64]): the item IDs the user has rated
        ratings(np.ndarray): the user's (normalized) ratings for those items
        other(np.ndarray): the item-feature matrix
        reg(float): the regularization term
    Returns:
        np.ndarray: the user-feature vector (equivalent to V in the current LU code)
    """
    # we can optimize by only considering the nonzero entries of Cu-I
    # this means we only need the corresponding matrix columns
    M = other[items, :]
    # Compute M^T (C_u-I) M, restricted to these nonzero entries
    MMT = (M.T.copy() * ratings) @ M
    # Build the matrix for solving
    A = otOr + MMT
    # Compute RHS - only used columns (p_ui != 0) values needed
    # Cu is rates + 1 for the cols, so just trim Ot
    y = other.T[:, items] @ (ratings + 1.0)
    # and solve
    _dposv(A, y, True)

    return y


class BiasedMF(MFPredictor):
    """
    Biased matrix factorization trained with alternating least squares [ZWSP2008]_.  This is a
    prediction-oriented algorithm suitable for explicit feedback data.

    It provides two solvers for the optimization step (the `method` parameter):

    ``'cd'`` (the default)
        Coordinate descent [TPT2011]_, adapted for a separately-trained bias model and to use
        weighted regularization as in the original ALS paper [ZWSP2008]_.
    ``'lu'``
        A direct implementation of the original ALS concept [ZWSP2008]_ using LU-decomposition
        to solve for the optimized matrices.

    See the base class :class:`.MFPredictor` for documentation on
    the estimated parameters you can extract from a trained model.

    .. [ZWSP2008] Yunhong Zhou, Dennis Wilkinson, Robert Schreiber, and Rong Pan. 2008.
        Large-Scale Parallel Collaborative Filtering for the Netflix Prize.
        In +Algorithmic Aspects in Information and Management_, LNCS 5034, 337–348.
        DOI `10.1007/978-3-540-68880-8_32 <http://dx.doi.org/10.1007/978-3-540-68880-8_32>`_.

    .. [TPT2011] Gábor Takács, István Pilászy, and Domonkos Tikk. 2011. Applications of the
        Conjugate Gradient Method for Implicit Feedback Collaborative Filtering.

    Args:
        features(int): the number of features to train
        iterations(int): the number of iterations to train
        reg(float): the regularization factor; can also be a tuple ``(ureg, ireg)`` to
            specify separate user and item regularization terms.
        damping(float): damping factor for the underlying bias.
        bias(bool or :class:`Bias`): the bias model.  If ``True``, fits a :class:`Bias` with
            damping ``damping``.
        method(str): the solver to use (see above).
        rng_spec:
            Random number generator or state (see :func:`lenskit.util.random.rng`).
        progress: a :func:`tqdm.tqdm`-compatible progress bar function
    """
    timer = None

    def __init__(self, features, *, iterations=20, reg=0.1, damping=5, bias=True, method='cd',
                 rng_spec=None, progress=None):
        self.features = features
        self.iterations = iterations
        self.regularization = reg
        self.damping = damping
        self.method = method
        if bias is True:
            self.bias = Bias(damping=damping)
        else:
            self.bias = bias
        self.progress = progress if progress is not None else util.no_progress
        self.rng = util.rng(rng_spec)

    def fit(self, ratings, **kwargs):
        """
        Run ALS to train a model.

        Args:
            ratings: the ratings data frame.

        Returns:
            The algorithm (for chaining).
        """

        self.timer = util.Stopwatch()

        for epoch, algo in enumerate(self.fit_iters(ratings, **kwargs)):
            pass  # we just need to do the iterations

        _logger.info('trained model in %s (|P|=%f, |Q|=%f)', self.timer,
                     np.linalg.norm(self.user_features_, 'fro'),
                     np.linalg.norm(self.item_features_, 'fro'))

        del self.timer
        return self

    def fit_iters(self, ratings, **kwargs):
        """
        Run ALS to train a model, returning each iteration as a generator.

        Args:
            ratings: the ratings data frame.

        Returns:
            The algorithm (for chaining).
        """

        if self.bias:
            _logger.info('[%s] fitting bias model', self.timer)
            self.bias.fit(ratings)

        current, uctx, ictx = self._initial_model(ratings)

        _logger.info('[%s] training biased MF model with ALS for %d features',
                     self.timer, self.features)
        for epoch, model in enumerate(self._train_iters(current, uctx, ictx)):
            self._save_params(model)
            yield self

    def _save_params(self, model):
        "Save the parameters into model attributes."
        self.item_index_ = model.items
        self.user_index_ = model.users
        self.item_features_ = model.item_matrix
        self.user_features_ = model.user_matrix

    def _initial_model(self, ratings, bias=None):
        # transform ratings using offsets
        if self.bias:
            ratings = self.bias.transform(ratings)

        "Initialize a model and build contexts."
        rmat, users, items = sparse_ratings(ratings)
        n_users = len(users)
        n_items = len(items)

        rmat = self._normalize(rmat, users, items)

        _logger.debug('setting up contexts')
        trmat = rmat.transpose()

        _logger.debug('initializing item matrix')
        imat = self.rng.standard_normal((n_items, self.features))
        imat /= np.linalg.norm(imat, axis=1).reshape((n_items, 1))
        _logger.debug('|Q|: %f', np.linalg.norm(imat, 'fro'))
        _logger.debug('initializing user matrix')
        umat = self.rng.standard_normal((n_users, self.features))
        umat /= np.linalg.norm(umat, axis=1).reshape((n_users, 1))
        _logger.debug('|P|: %f', np.linalg.norm(umat, 'fro'))

        return PartialModel(users, items, umat, imat), rmat, trmat

    def _normalize(self, ratings, users, items):
        "Apply bias normalization to the data in preparation for training."
        n_users = len(users)
        n_items = len(items)
        assert ratings.nrows == n_users
        assert ratings.ncols == n_items

        if self.bias:
            if ratings.values is None:
                ratings.N.values = np.ones(np.nnz)

        _logger.info('[%s] normalizing %dx%d matrix (%d nnz)',
                     self.timer, n_users, n_items, ratings.nnz)

        return ratings

    def _train_iters(self, current, uctx, ictx):
        """
        Generator of training iterations.

        Args:
            current(PartialModel): the current model step.
            uctx(ndarray): the user-item rating matrix for training user features.
            ictx(ndarray): the item-user rating matrix for training item features.
        """
        n_items = len(current.items)
        n_users = len(current.users)
        assert uctx.nrows == n_users
        assert uctx.ncols == n_items
        assert ictx.nrows == n_items
        assert ictx.ncols == n_users

        if self.method == 'cd':
            train = _train_matrix_cd
        elif self.method == 'lu':
            train = _train_matrix_lu
        else:
            raise ValueError('invalid training method ' + self.method)

        if isinstance(self.regularization, tuple):
            ureg, ireg = self.regularization
        else:
            ureg = ireg = self.regularization

        for epoch in self.progress(range(self.iterations), desc='BiasedMF', leave=False):
            du = train(uctx.N, current.user_matrix, current.item_matrix, ureg)
            _logger.debug('[%s] finished user epoch %d', self.timer, epoch)
            di = train(ictx.N, current.item_matrix, current.user_matrix, ireg)
            _logger.debug('[%s] finished item epoch %d', self.timer, epoch)
            _logger.info('[%s] finished epoch %d (|ΔP|=%.3f, |ΔQ|=%.3f)', self.timer, epoch, du, di)
            yield current

    def predict_for_user(self, user, items, ratings=None):
        scores = None
        u_offset = None
        if ratings is not None and len(ratings) > 0:
            if self.bias:
                ratings, u_offset = self.bias.transform_user(ratings)

            ri_idxes = self.item_index_.get_indexer_for(ratings.index)
            ri_good = ri_idxes >= 0
            ri_it = ri_idxes[ri_good]
            ri_val = ratings.values[ri_good]

            # unpack regularization
            if isinstance(self.regularization, tuple):
                ureg, ireg = self.regularization
            else:
                ureg = ireg = self.regularization

            u_feat = _train_bias_row_lu(ri_it, ri_val, self.item_features_, ureg)
            scores = self.score_by_ids(user, items, u_feat)
        else:
            # look up user index
            scores = self.score_by_ids(user, items)

        if self.bias and ratings is not None and len(ratings) > 0:
            return self.bias.inverse_transform_user(user, scores, u_offset)
        elif self.bias:
            return self.bias.inverse_transform_user(user, scores)
        else:
            return scores

    def __str__(self):
        return 'als.BiasedMF(features={}, regularization={})'.\
            format(self.features, self.regularization)


class ImplicitMF(MFPredictor):
    """
    Implicit matrix factorization trained with alternating least squares [HKV2008]_.  This
    algorithm outputs 'predictions', but they are not on a meaningful scale.  If its input
    data contains ``rating`` values, these will be used as the 'confidence' values; otherwise,
    confidence will be 1 for every rated item.

    ``'cg'`` (the default)
        Conjugate gradient method [TPT2011]_.
    ``'lu'``
        A direct implementation of the original implicit-feedback ALS concept [HKV2008]_ using
        LU-decomposition to solve for the optimized matrices.

    See the base class :class:`.MFPredictor` for documentation on the estimated parameters
    you can extract from a trained model.

    .. [HKV2008] Y. Hu, Y. Koren, and C. Volinsky. 2008.
       Collaborative Filtering for Implicit Feedback Datasets.
       In _Proceedings of the 2008 Eighth IEEE International Conference on Data Mining_, 263–272.
       DOI `10.1109/ICDM.2008.22 <http://dx.doi.org/10.1109/ICDM.2008.22>`_

    .. [TPT2011] Gábor Takács, István Pilászy, and Domonkos Tikk. 2011. Applications of the
        Conjugate Gradient Method for Implicit Feedback Collaborative Filtering.

    Args:
        features(int): the number of features to train
        iterations(int): the number of iterations to train
        reg(double): the regularization factor
        weight(double): the scaling weight for positive samples (:math:`\\alpha` in [HKV2008]_).
        rng_spec:
            Random number generator or state (see :func:`lenskit.util.random.rng`).
        progress: a :func:`tqdm.tqdm`-compatible progress bar function
    """
    timer = None

    def __init__(self, features, *, iterations=20, reg=0.1, weight=40, method='cg',
                 rng_spec=None, progress=None):
        self.features = features
        self.iterations = iterations
        self.reg = reg
        self.weight = weight
        self.method = method
        self.rng = util.rng(rng_spec)
        self.progress = progress if progress is not None else util.no_progress

    def fit(self, ratings, **kwargs):
        self.timer = util.Stopwatch()
        for algo in self.fit_iters(ratings, **kwargs):
            pass

        _logger.info('[%s] finished training model with %d features (|P|=%f, |Q|=%f)',
                     self.timer, self.features,
                     np.linalg.norm(self.user_features_, 'fro'),
                     np.linalg.norm(self.item_features_, 'fro'))

        # unpack the regularization
        if isinstance(self.reg, tuple):
            ureg, ireg = self.reg
        else:
            ureg = ireg = self.reg

        # compute OtOr and save it on the model
        self.OtOr_ = _implicit_otor(self.item_features_, ureg)

        return self

    def fit_iters(self, ratings, **kwargs):
        current, uctx, ictx = self._initial_model(ratings)

        _logger.info('[%s] training implicit MF model with ALS for %d features',
                     self.timer, self.features)
        _logger.info('have %d observations for %d users and %d items',
                     uctx.nnz, uctx.nrows, ictx.nrows)
        for model in self._train_iters(current, uctx, ictx):
            self._save_model(model)
            yield self

    def _save_model(self, model):
        self.item_index_ = model.items
        self.user_index_ = model.users
        self.item_features_ = model.item_matrix
        self.user_features_ = model.user_matrix

    def _train_iters(self, current, uctx, ictx):
        "Generator of training iterations."
        if self.method == 'lu':
            train = _train_implicit_lu
        elif self.method == 'cg':
            train = _train_implicit_cg
        else:
            raise ValueError('unknown solver ' + self.method)

        if isinstance(self.reg, tuple):
            ureg, ireg = self.reg
        else:
            ureg = ireg = self.reg

        for epoch in self.progress(range(self.iterations), desc='ImplicitMF', leave=False):
            du = train(uctx.N, current.user_matrix, current.item_matrix, ureg)
            _logger.debug('[%s] finished user epoch %d', self.timer, epoch)
            di = train(ictx.N, current.item_matrix, current.user_matrix, ireg)
            _logger.debug('[%s] finished item epoch %d', self.timer, epoch)
            _logger.info('[%s] finished epoch %d (|ΔP|=%.3f, |ΔQ|=%.3f)', self.timer, epoch, du, di)
            yield current

    def _initial_model(self, ratings):
        "Initialize a model and build contexts."

        rmat, users, items = sparse_ratings(ratings)
        n_users = len(users)
        n_items = len(items)

        _logger.debug('setting up contexts')
        # force values to exist
        if rmat.values is None:
            rmat.values = np.ones(rmat.nnz)
        rmat.values *= self.weight
        trmat = rmat.transpose()

        imat = self.rng.standard_normal((n_items, self.features)) * 0.01
        imat = np.square(imat)
        umat = self.rng.standard_normal((n_users, self.features)) * 0.01
        umat = np.square(umat)

        return PartialModel(users, items, umat, imat), rmat, trmat

    def predict_for_user(self, user, items, ratings=None):
        if ratings is not None and len(ratings) > 0:
            ri_idxes = self.item_index_.get_indexer_for(ratings.index)
            ri_good = ri_idxes >= 0
            ri_it = ri_idxes[ri_good]
            ri_val = ratings.values[ri_good]
            ri_val *= self.weight
            u_feat = _train_implicit_row_lu(ri_it, ri_val, self.item_features_, self.OtOr_)
            return self.score_by_ids(user, items, u_feat)
        else:
            # look up user index
            return self.score_by_ids(user, items)

    def __str__(self):
        return 'als.ImplicitMF(features={}, reg={}, w={})'.\
            format(self.features, self.reg, self.weight)



"""
Logging utilities.
"""

import sys
import logging
from logging.handlers import QueueListener
import multiprocessing as mp

_log = logging.getLogger(__name__)
_lts_initialized = False
_ltn_initialized = False
_log_queue = None
_log_listener = None


class InjectHandler:
    "Handler that re-injects a message into parent process logging"
    level = logging.DEBUG

    def handle(self, record):
        logger = logging.getLogger(record.name)
        if logger.isEnabledFor(record.levelno):
            logger.handle(record)


class LowPassFilter:
    def filter(record):
        return record.levelno < logging.WARNING


def log_to_stderr(level=logging.INFO):
    """
    Set up the logging infrastructure to show log output on ``sys.stderr``, where it will
    appear in the IPython message log.
    """
    global _lts_initialized
    if _lts_initialized:
        _log.info('log already initialized')

    h = logging.StreamHandler(sys.stderr)
    f = logging.Formatter('[%(levelname)7s] %(name)s %(message)s')
    h.setFormatter(f)
    root = logging.getLogger()
    root.addHandler(h)
    root.setLevel(level)

    _log.info('stderr logging configured')
    _lts_initialized = True


def log_to_notebook(level=logging.INFO):
    """
    Set up the logging infrastructure to show log output in the Jupyter notebook.
    """
    global _ltn_initialized
    if _ltn_initialized:
        _log.info('log already initialized')

    h = logging.StreamHandler(sys.stderr)
    f = logging.Formatter('[%(levelname)7s] %(name)s %(message)s')
    h.setFormatter(f)
    h.setLevel(logging.WARNING)

    oh = logging.StreamHandler(sys.stdout)
    oh.setFormatter(f)
    oh.addFilter(LowPassFilter)

    root = logging.getLogger()
    root.addHandler(h)
    root.addHandler(oh)
    root.setLevel(level)

    _log.info('notebook logging configured')
    _ltn_initialized = True


def log_queue():
    """
    Get the log queue for child process logging.
    """
    global _log_queue, _log_listener
    from lenskit.util.parallel import LKContext
    ctx = LKContext.INSTANCE
    if _log_queue is None:
        _log_queue = ctx.Queue()
        _log_listener = QueueListener(_log_queue, InjectHandler())
        _log_listener.start()
    return _log_queue



"""
Miscellaneous utility functions.
"""

import logging
from copy import deepcopy

from ..algorithms import Algorithm
from .log import log_to_notebook, log_to_stderr  # noqa: F401
from .timing import Stopwatch  # noqa: F401
from .data import read_df_detect  # noqa: F401
from .random import rng, init_rng, derivable_rng  # noqa: F401
from .parallel import proc_count  # noqa: F401

try:
    import resource
except ImportError:
    resource = None

_log = logging.getLogger(__name__)

__all__ = [
    'log_to_stderr', 'log_to_notebook',
    'Stopwatch',
    'read_df_detect',
    'rng', 'init_rng', 'derivable_rng',
    'proc_count',
    'clone'
]


def clone(algo):
    """
    Clone an algorithm, but not its fitted data.  This is like
    :py:func:`scikit.base.clone`, but may not work on arbitrary SciKit estimators.
    LensKit algorithms are compatible with SciKit clone, however, so feel free
    to use that if you need more general capabilities.

    This function is somewhat derived from the SciKit one.

    >>> from lenskit.algorithms.bias import Bias
    >>> orig = Bias()
    >>> copy = clone(orig)
    >>> copy is orig
    False
    >>> copy.damping == orig.damping
    True
    """
    _log.debug('cloning %s', algo)
    if isinstance(algo, Algorithm) or hasattr(algo, 'get_params'):
        params = algo.get_params(deep=False)

        sps = dict([(k, clone(v)) for (k, v) in params.items()])
        return algo.__class__(**sps)
    elif isinstance(algo, list) or isinstance(algo, tuple):
        return [clone(a) for a in algo]
    else:
        return deepcopy(algo)


class LastMemo:
    def __init__(self, func, check_type='identity'):
        self.function = func
        self.check = check_type
        self.memory = None
        self.result = None

    def __call__(self, arg):
        if not self._arg_is_last(arg):
            self.result = self.function(arg)
            self.memory = arg

        return self.result

    def _arg_is_last(self, arg):
        if self.check == 'identity':
            return arg is self.memory
        elif self.check == 'equality':
            return arg == self.memory


def last_memo(func=None, check_type='identity'):
    if func is None:
        return lambda f: LastMemo(f, check_type)
    else:
        return LastMemo(func, check_type)


def no_progress(obj, **kwargs):
    return obj


def max_memory():
    "Get the maximum memory use for this process"
    if resource:
        res = resource.getrusage(resource.RUSAGE_SELF)
        return "%.1f MiB" % (res.ru_maxrss,)
    else:
        return 'unknown'


def cur_memory():
    "Get the current memory use for this process"
    if resource:
        res = resource.getrusage(resource.RUSAGE_SELF)
        return "%.1f MiB" % (res.ru_idrss,)
    else:
        return 'unknown'



import logging
import itertools as it
import warnings
from collections import OrderedDict as od

import numpy as np
import pandas as pd

from .metrics.topn import *
from .util import Stopwatch
from .util.parallel import invoker

_log = logging.getLogger(__name__)


def _length(df, *args, **kwargs):
    return float(len(df))


def _grouping_iter(df, cols, ksf=()):
    # what key are we going to work on?
    current = cols[0]
    remaining = cols[1:]

    if not remaining:
        # this is the last column. we group by it.
        for gk, gdf in df.groupby(current):
            yield ksf + (gk,), gdf.drop(columns=[current])
    else:
        # we have columns remaining, let's start grouping by this one
        for v, subdf in df.groupby(current):
            yield from _grouping_iter(subdf.drop(columns=[current]), remaining, ksf + (v,))


def _rla_worker(model, req):
    col, val = req
    return model._compute_group(col, val)


class _RLAJob:
    def __init__(self, recs, truth, metrics):
        self.recs = recs
        self.truth = truth
        self.metrics = metrics

    def prepare(self, group_cols):
        rec_key, truth_key = _df_keys(self.recs.columns, self.truth.columns, group_cols)
        _log.info('collecting truth data')
        truth_frames = dict((k, df.set_index('item'))
                            for (k, df)
                            in _grouping_iter(self.truth, truth_key))
        _log.debug('found truth for %d users', len(truth_frames))
        self.truth = truth_frames
        self.rec_key = rec_key
        self.truth_key = truth_key

    def compute(self, n_jobs=None):
        first = self.rec_key[0]
        bins = self.recs.groupby(first)['item'].count()
        total = bins.sum()
        _log.debug('info RLA for %d rows in %d bins', total, len(bins))
        if total < 1000 or len(bins) < 4:
            n_jobs = 1  # force in-process for small runs

        with invoker(self, _rla_worker, n_jobs) as loop:
            res = loop.map((first, v) for v in bins.index.values)
            res = pd.concat(res, ignore_index=True)

        return res.set_index(self.rec_key)

    def _compute_group(self, col, val):
        _log.debug('computing for %s=%s', col, val)
        mnames = [mn for (mf, mn, margs) in self.metrics]
        gen = self._iter_measurements(col, val)
        return pd.DataFrame.from_records(gen, columns=self.rec_key + mnames)

    def _iter_measurements(self, col, val):
        key = self.rec_key[1:]
        df = self.recs[self.recs[col] == val]
        df = df.drop(columns=[col])
        nt = len(self.truth_key)
        if key:
            for rk, gdf in _grouping_iter(df, key):
                rk = (val,) + rk
                tk = rk[-nt:]
                g_truth = self.truth[tk]
                results = tuple(mf(gdf, g_truth, **margs) for (mf, mn, margs) in self.metrics)
                yield rk + results
        else:
            # we only have one group level
            tk = (val,)
            g_truth = self.truth[tk]
            results = tuple(mf(df, g_truth, **margs) for (mf, mn, margs) in self.metrics)
            yield tk + results


class RecListAnalysis:
    """
    Compute one or more top-N metrics over recommendation lists.

    This method groups the recommendations by the specified columns,
    and computes the metric over each group.  The default set of grouping
    columns is all columns *except* the following:

    * ``item``
    * ``rank``
    * ``score``
    * ``rating``

    The truth frame, ``truth``, is expected to match over (a subset of) the
    grouping columns, and contain at least an ``item`` column.  If it also
    contains a ``rating`` column, that is used as the users' rating for
    metrics that require it; otherwise, a rating value of 1 is assumed.

    .. warning::
       Currently, RecListAnalysis will silently drop users who received
       no recommendations.  We are working on an ergonomic API for fixing
       this problem.

    Args:
        group_cols(list):
            The columns to group by, or ``None`` to use the default.
    """

    DEFAULT_SKIP_COLS = ['item', 'rank', 'score', 'rating']

    def __init__(self, group_cols=None, n_jobs=None):
        self.group_cols = group_cols
        self.metrics = [(_length, 'nrecs', {})]
        self.n_jobs = n_jobs

    def add_metric(self, metric, *, name=None, **kwargs):
        """
        Add a metric to the analysis.

        A metric is a function of two arguments: the a single group of the recommendation
        frame, and the corresponding truth frame.  The truth frame will be indexed by
        item ID.  The recommendation frame will be in the order in the data.  Many metrics
        are defined in :mod:`lenskit.metrics.topn`; they are re-exported from
        :mod:`lenskit.topn` for convenience.

        Args:
            metric: The metric to compute.
            name: The name to assign the metric. If not provided, the function name is used.
            **kwargs: Additional arguments to pass to the metric.
        """
        if name is None:
            name = metric.__name__

        self.metrics.append((metric, name, kwargs))

    def compute(self, recs, truth, *, include_missing=False):
        """
        Run the analysis.  Neither data frame should be meaningfully indexed.

        Args:
            recs(pandas.DataFrame):
                A data frame of recommendations.
            truth(pandas.DataFrame):
                A data frame of ground truth (test) data.
            include_missing(bool):
                ``True`` to include users from truth missing from recs.
                Matches are done via group columns that appear in both
                ``recs`` and ``truth``.

        Returns:
            pandas.DataFrame: The results of the analysis.
        """
        _log.info('analyzing %d recommendations (%d truth rows)', len(recs), len(truth))

        job = _RLAJob(recs, truth, self.metrics)
        job.prepare(self.group_cols)

        timer = Stopwatch()
        _log.info('collecting metric results')
        res = job.compute(self.n_jobs)
        _log.info('measured %d lists in %s', len(res), timer)

        if include_missing:
            _log.info('filling in missing user info')
            ug_cols = [c for c in job.rec_key if c not in job.truth_key]
            tcount = truth.reset_index().groupby(job.truth_key)['item'].count()
            tcount.name = 'ntruth'
            _log.debug('res index levels: %s', res.index.names)
            if ug_cols:
                _log.debug('regrouping by %s to fill', ug_cols)
                res = res.groupby(level=ug_cols).apply(lambda f: f.reset_index(ug_cols, drop=True).join(tcount, how='outer'))
            else:
                _log.debug('no ungroup cols, directly merging to fill')
                res = res.join(tcount, how='outer')
            _log.debug('final columns: %s', res.columns)
            _log.debug('index levels: %s', res.index.names)
            res['ntruth'] = res['ntruth'].fillna(0)
            res['nrecs'] = res['nrecs'].fillna(0)

        return res


def _df_keys(r_cols, t_cols, g_cols=None, skip_cols=RecListAnalysis.DEFAULT_SKIP_COLS):
    "Identify rec list and truth list keys."
    if g_cols is None:
        g_cols = [c for c in r_cols if c not in skip_cols]

    truth_key = [c for c in g_cols if c in t_cols]
    rec_key = [c for c in g_cols if c not in t_cols] + truth_key
    _log.info('using rec key columns %s', rec_key)
    _log.info('using truth key columns %s', truth_key)
    return rec_key, truth_key
